from rest_framework.serializers import ModelSerializer


def get_serializer(model):
    if model.get_serializer(model) is not None:
        return model.get_serializer(model)
    return type(f'{model.__name__}Serializer', (ModelSerializer, ), {
        'Meta': type('Meta', (object,), {
            'model': model,
            'fields': '__all__'
        })
    })



from .base_page import BasePage
from django.db.models import Q
from garpix_utils.paginator import GarpixPaginator


class BaseSearchPage(BasePage):
    paginate_by = 25
    template = 'garpix_page/default_search.html'

    def get_context(self, request=None, *args, **kwargs):
        from ..utils.get_garpix_page_models import get_garpix_page_models
        context = super().get_context(request, *args, **kwargs)
        search_query = request.GET.get('q', None)
        object_list = BasePage.objects.none()

        try:
            page = int(request.GET.get('page', 1))
        except ValueError:
            page = 1

        if search_query is not None:
            ids = []
            for Model in get_garpix_page_models():
                q = Q()
                for searchable_field in Model.searchable_fields:
                    searchable_dict = {
                        f'{searchable_field}__icontains': search_query,
                    }
                    q |= Q(**searchable_dict)
                ids += list(Model.on_site.filter(is_active=True).filter(q).values_list('id', flat=True))
            object_list = BasePage.objects.filter(id__in=ids)

        paginator = GarpixPaginator(object_list, self.paginate_by)
        paginated_object_list = paginator.get_page(page)
        context.update({
            'paginator': paginator,
            'paginated_object_list': paginated_object_list,
            'page': page,
            'q': search_query or '',
        })
        return context

    class Meta:
        abstract = True



from django.contrib import admin
from django.contrib.admin.widgets import FilteredSelectMultiple
from django.utils.text import format_lazy
from modeltranslation.admin import TabbedTranslationAdmin
from polymorphic.admin import PolymorphicParentModelAdmin, PolymorphicChildModelFilter, PolymorphicChildModelAdmin

from garpix_page.models import BaseComponent, BasePage
from garpix_page.utils.get_garpix_page_models import get_garpix_page_component_models


class BaseComponentAdmin(PolymorphicChildModelAdmin, TabbedTranslationAdmin):
    base_model = BaseComponent
    list_display = ('title', 'model_name')

    filter_horizontal = (
        'pages',
    )

    def formfield_for_manytomany(self, db_field, request, **kwargs):

        if db_field.name == 'pages':
            kwargs['widget'] = FilteredSelectMultiple(
                db_field.verbose_name, is_stacked=False
            )
        else:
            return super().formfield_for_manytomany(db_field, request, **kwargs)
        if 'queryset' not in kwargs:
            queryset = BasePage.objects.all()
            if queryset is not None:
                kwargs['queryset'] = queryset
        form_field = db_field.formfield(**kwargs)
        msg = 'Hold down “Control”, or “Command” on a Mac, to select more than one.'
        help_text = form_field.help_text
        form_field.help_text = (
            format_lazy('{} {}', help_text, msg) if help_text else msg
        )
        return form_field


@admin.register(BaseComponent)
class RealBaseComponentAdmin(PolymorphicParentModelAdmin, TabbedTranslationAdmin):

    child_models = get_garpix_page_component_models()
    base_model = BaseComponent
    list_filter = (PolymorphicChildModelFilter, 'pages')
    list_display = ('title', 'model_name')



from .base_page import BasePage
from garpix_utils.paginator import GarpixPaginator


class BaseListPage(BasePage):
    paginate_by = 25
    template = 'garpix_page/default_list.html'

    def get_context(self, request=None, *args, **kwargs):
        context = super().get_context(request, *args, **kwargs)

        object_list = self.children.filter(is_active=True)
        paginator = GarpixPaginator(object_list, self.paginate_by)

        try:
            page = int(request.GET.get('page', 1))
        except ValueError:
            page = 1

        paginated_object_list = paginator.get_page(page)

        context.update({
            'paginator': paginator,
            'paginated_object_list': paginated_object_list,
            'page': page,
        })
        return context

    class Meta:
        abstract = True



from django.utils import translation
from django.conf import settings


def get_current_language_code_url_prefix():
    current_language_code_url_prefix = translation.get_language()
    try:
        use_default_prefix = settings.USE_DEFAULT_LANGUAGE_PREFIX
    except:  # noqa
        use_default_prefix = True
    if not use_default_prefix and current_language_code_url_prefix == settings.LANGUAGE_CODE:
        current_language_code_url_prefix = ''
    elif current_language_code_url_prefix is None:
        current_language_code_url_prefix = ''
    else:
        current_language_code_url_prefix = '/' + current_language_code_url_prefix
    return current_language_code_url_prefix



from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sites.models import Site
from rest_framework.test import APIClient, APITestCase
from model_bakery import baker

from ..utils.get_garpix_page_models import get_garpix_page_models


class BasePageApiTest(APITestCase):

    @classmethod
    def setUpClass(cls):
        cls.update_baker_default_mapping()
        super().setUpClass()

    def setUp(self):
        self.client = APIClient()
        page_models = get_garpix_page_models()
        sites = Site.objects.all()

        self.pages = []
        i = 0
        for page_model in page_models:
            page = baker.make(page_model, slug=f'slug{i}', sites=sites)
            self.pages.append(page)
            i += 1
        if len(page_models) > 0:
            page = baker.make(page_models[0], slug='', sites=sites)
            self.pages.append(page)
        self.test_user = baker.make(get_user_model())
        self.languages_list = [x[0] for x in settings.LANGUAGES]

    def test_page(self):
        for page in self.pages:
            response = self.client.get(f'/{page.slug}')
            if getattr(page, 'login_required', False):
                self.assertEqual(response.status_code, 302, f'Error in page {page} ({page.model_name()})')
                self.user_login()
                response = self.client.get(f'/{page.slug}')
            if not page.has_permission_required(response.wsgi_request):
                self.assertEqual(response.status_code, 302, f'Error in page {page} ({page.model_name()})')
            else:
                self.assertNotRegex(str(response.status_code), r'^5\d{2}$',
                                    f'Error in page {page} ({page.model_name()})')
            self.client.logout()

    def test_page_api(self):
        if hasattr(settings, 'API_URL'):
            for page in self.pages:
                responses = self.generate_responses_list(page)
                if getattr(page, 'login_required', False):
                    self.check_response_status(responses, 401)
                    self.user_login()
                    responses = self.generate_responses_list(page)
                if not page.has_permission_required(responses[0][0].wsgi_request):
                    self.user_login()
                    responses = self.generate_responses_list(page)
                    self.check_response_status(responses, 403)
                else:
                    self.check_response_status(responses)
                self.client.logout()

    def user_login(self):
        self.client.force_login(self.test_user)
        self.client.force_authenticate(self.test_user)

    def generate_responses_list(self, page):
        responses = [
            (self.client.get(f'/{settings.API_URL}/page/{page.slug}'), page),
            (self.client.get(f'/{settings.API_URL}/page/{page.slug}/'), page)
        ]
        for language in self.languages_list:
            responses.append((self.client.get(f'/{settings.API_URL}/page/{language}/{page.slug}'), page))
            responses.append((self.client.get(f'/{settings.API_URL}/page/{language}/{page.slug}/'), page))
        return responses

    def check_response_status(self, responses, status_code=None):
        if not status_code:
            for response in responses:
                self.assertNotRegex(str(response[0].status_code), r'^5\d{2}$',
                                    f'Error in page api of {response[1]} ({response[1].model_name()})')
        else:
            for response in responses:
                self.assertEqual(response[0].status_code, status_code,
                                 f'Error in page api of {response[1]} ({response[1].model_name()})')

    @staticmethod
    def update_baker_default_mapping():

        def gen_phone_number() -> str:
            import random
            import string
            return f"+{''.join(random.SystemRandom().choice(string.digits) for _ in range(15))}"

        from model_bakery import random_gen
        from model_bakery.generators import default_mapping

        try:
            from ckeditor.fields import RichTextField
        except ImportError:
            RichTextField = None

        try:
            from ckeditor_uploader.fields import RichTextUploadingField
        except ImportError:
            RichTextUploadingField = None

        try:
            from phonenumber_field.modelfields import PhoneNumberField
        except ImportError:
            PhoneNumberField = None

        default_mapping.update({
            RichTextField: random_gen.gen_text,
            RichTextUploadingField: random_gen.gen_text,
            PhoneNumberField: gen_phone_number
        })



from rest_framework import status
from django.utils.translation import activate
from rest_framework import views
from rest_framework.response import Response
from django.contrib.auth import get_user_model
import django.apps
from django.utils.module_loading import import_string
from django.conf import settings
from ..serializers.serializer import get_serializer

from ..utils.get_current_language_code_url_prefix import get_current_language_code_url_prefix

model_list = []
for model in django.apps.apps.get_models():
    try:
        if model.is_for_page_view():
            model_list.append(model)
    except:  # noqa
        pass

languages_list = [x[0] for x in settings.LANGUAGES]


class PageApiView(views.APIView):

    @staticmethod
    def get_absolute_url_from_request(slug_list, slug):
        current_language_code_url_prefix = get_current_language_code_url_prefix()
        if slug != '':
            return "{}/{}".format(current_language_code_url_prefix, '/'.join(slug_list))
        return "{}".format(current_language_code_url_prefix) if len(current_language_code_url_prefix) > 1 else '/'

    @staticmethod
    def get_instance_by_slug(slug):
        for m in model_list:
            instance = m.objects.filter(slug=slug).first()
            if instance:
                return instance
        return None

    def get_object(self, slug):
        obj = self.get_instance_by_slug(slug)
        return obj

    @staticmethod
    def get_error_page_response_data(page, request):
        return {
            'page_model': page.__class__.__name__ if page is not None else None,
            'init_state': {
                "object": None,
                "global": import_string(settings.GARPIX_PAGE_GLOBAL_CONTEXT)(request, page)
            }
        }

    def check_errors(self, page, request):
        if page is None:
            return Response(self.get_error_page_response_data(page, request), status=status.HTTP_404_NOT_FOUND)

        if getattr(page, 'login_required', False):
            if not request.user.is_authenticated:
                return Response(self.get_error_page_response_data(page, request), status=status.HTTP_401_UNAUTHORIZED)

        if not page.has_permission_required(request):
            return Response(self.get_error_page_response_data(page, request), status=status.HTTP_403_FORBIDDEN)

        if getattr(page, 'query_parameters_required', None) is not None:
            request_get = set(request.GET.keys())
            parameters = set(page.query_parameters_required)
            if request_get != parameters:
                return Response(self.get_error_page_response_data(page, request), status=status.HTTP_404_NOT_FOUND)

        return None

    def get(self, request, slugs):  # noqa

        language = languages_list[0]
        if 'HTTP_ACCEPT_LANGUAGE' in request.META and request.META['HTTP_ACCEPT_LANGUAGE'] in languages_list:
            language = request.META['HTTP_ACCEPT_LANGUAGE']
        activate(language)

        slug_list = slugs.split('/')

        if slug_list[0] in languages_list:
            activate(slug_list[0])
            slug_list.pop(0)

        if len(slug_list) == 0:
            slug_list.append('')

        if len(slug_list) > 1 and slug_list[-1] == '':
            slug_list.pop(-1)

        slug = slug_list[-1]
        page = self.get_object(slug)

        if page and page.absolute_url != self.get_absolute_url_from_request(slug_list, slug):
            page = None

        errors = self.check_errors(page, request)
        if errors is not None:
            return errors

        if request.user.is_authenticated:
            user = get_user_model().objects.get(pk=request.user.pk)
        else:
            user = None

        page_context = page.get_context(request, object=page, user=user)
        request = page_context.pop('request')
        for k, v in page_context.items():
            if hasattr(v, 'is_for_page_view'):
                model_serializer_class = get_serializer(v.__class__)
                page_context[k] = model_serializer_class(v, context={"request": request}).data
        if 'paginated_object_list' in page_context:
            if page_context['paginated_object_list'][0].get_serializer() is not None:
                page_context['paginated_object_list'] = page_context['paginated_object_list'][0].get_serializer(
                    page_context['paginated_object_list'], context={"request": request}, many=True).data
            else:
                page_context['paginated_object_list'] = list(
                    {'id': x.id, 'title': x.title, 'get_absolute_url': x.get_absolute_url()} for x in
                    page_context['paginated_object_list'])
        if 'paginator' in page_context:
            page_context['num_pages'] = page_context['paginator'].num_pages
            page_context['per_page'] = page_context['paginator'].per_page
            page_context.pop('paginator')

        page_context['global'] = import_string(settings.GARPIX_PAGE_GLOBAL_CONTEXT)(request, page)
        page_context['object'].update({
            'components': page.get_components_context(request)
        })
        data = {
            'page_model': page.__class__.__name__,
            'init_state': page_context,
        }
        return Response(data)
